/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */
/*
if_idx_nextop.c
*/
#define VERSION "Version 0.05 -- 000329"


/*

 This is a utility function for IP-login probing. It uses the Linux
 netlink feature. Netlink is a linux "development" of routing socket.
 It looks up the outgoing interface index based on the destination
 IP address. It parses the unicast routing table for longest match.

 Computer Dept.
 P.O Box 7079
 S-75107 Uppsala 
 Sweden.
  
 Robert Olsson <robert.olsson@data.slu.se>


This work heavily relies on GPL'ed work with the following copyright notice:

1> libnetlink.c RTnetlink service routines
1> Authors:     Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>

2> rt_netlink.c
2> Copyright (C) 1997, 98, 99 Kunihiro Ishiguro

*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <errno.h>

#include "if_idx_nexthop.h"

/* Global vars */

/* The IP address we have */
struct in_addr match;

/* The index we looking for */
int if_index;

struct best_route best_route;

struct 
{
  int sock;
  int seq;
  struct sockaddr_nl snl;
} netlink = { -1, 0, {0} };

typedef struct
{
        __u8 family;
        __u8 bytelen;
        __s16 bitlen;
        __u32 data[4];
} inet_prefix;

/* Utility function for parse rtattr. */
static void
netlink_parse_rtattr (struct rtattr **tb, int max, struct rtattr *rta, int len)
{
  while (RTA_OK(rta, len)) 
    {
      if (rta->rta_type <= max)
	tb[rta->rta_type] = rta;
      rta = RTA_NEXT(rta,len);
    }
}

int inet_addr_match(inet_prefix *a, inet_prefix *b, int bits)
{
        __u32 *a1 = a->data;
        __u32 *a2 = b->data;
        int words = bits >> 0x05;

	/* Return 0 if address matches */

	/*
	printf(" *a1 %s \n",  inet_ntoa(*(struct in_addr *)a1));
	printf(" *a2 %s \n",  inet_ntoa(*(struct in_addr *)a2));
	*/

        bits &= 0x1f;
        if (words)
                if (memcmp(a1, a2, words << 2))
                        return -1;
        if (bits) {
                __u32 w1, w2;
                __u32 mask;

                w1 = a1[words];
                w2 = a2[words];
                mask = htonl((0xffffffff) << (0x20 - bits));
                if ( (w1 ^ w2) & mask)
                        return 1;
        }
        return 0;
}

/* Looking up routing table by netlink interface. */
int
netlink_routing_table (struct sockaddr_nl *snl, struct nlmsghdr *h)
{
  int len;
  struct rtmsg *rtm;
  struct rtattr *tb [RTA_MAX + 1];
  struct in_addr dest;
  struct in_addr gate;
  char anyaddr[16] = {0};
  int table;
  inet_prefix m, d;
  int index;

  if(if_index >=  0)  return 0; 

  rtm = NLMSG_DATA (h);
  if (h->nlmsg_type != RTM_NEWROUTE) return 0;
  len = h->nlmsg_len - NLMSG_LENGTH(sizeof (struct rtmsg));
  if (len < 0) return -1;

  if (rtm->rtm_family != AF_INET) return 0;
  if (rtm->rtm_type != RTN_UNICAST) return 0;
  if (rtm->rtm_flags & RTM_F_CLONED) return 0;
  if (rtm->rtm_protocol == RTPROT_REDIRECT)  return 0;
  /*  if (rtm->rtm_protocol == RTPROT_KERNEL) return 0; */
  if (rtm->rtm_src_len != 0) return -1;

  table = rtm->rtm_table;
#if 0		/* we weed them out later in rib_weed_tables () */
  if (table != RT_TABLE_MAIN && table != RTM_TABLE_DEFAULT)return 0;
#endif

  bzero (tb, sizeof tb);
  netlink_parse_rtattr (tb, RTA_MAX, RTM_RTA (rtm), len);
  
  bzero (&dest, sizeof dest);
  bzero (&gate, sizeof gate);

  if (tb[RTA_OIF]) index = *(int *) RTA_DATA (tb[RTA_OIF]);
  else return 0;

  if (tb[RTA_DST]) {
    dest.s_addr = *( unsigned long int *)  RTA_DATA (tb[RTA_DST]);
  }
  else
    dest.s_addr = *( unsigned long int *)  anyaddr;
  

  if (tb[RTA_GATEWAY]) {
    gate.s_addr =  *( unsigned long int *) RTA_DATA (tb[RTA_GATEWAY]);
  }
#if 0
  printf(" Matching %s ",  inet_ntoa(match));
  printf(" dest=%s, gateway=%s\n", inet_ntoa(dest), inet_ntoa(gate));
  printf(" %s/%d,  %d\n", inet_ntoa(dest), rtm->rtm_dst_len, index);
#endif 

  memcpy(m.data, &match.s_addr,  4);
  memcpy(d.data, &dest.s_addr, 4);

  if(inet_addr_match(&d, &m, rtm->rtm_dst_len) ==  0 ) {
    best_route.idx = if_index = index;
    best_route.gw =  gate.s_addr; 
    best_route.dst =  dest.s_addr; 
    best_route.mlen =  rtm->rtm_dst_len;
    return 1;
  }
  return 0;
}

/* Get type specified information from netlink. */
int
netlink_request (int family, int type)
{
  int ret;
  struct sockaddr_nl snl;

  struct
  {
    struct nlmsghdr nlh;
    struct rtgenmsg g;
  } req;

  /* Check netlink socket. */
  if (netlink.sock < 0)
    {
      perror( "netlink socket isn't active.");
      exit (-1);
    }

  bzero (&snl, sizeof snl);
  snl.nl_family = AF_NETLINK;

  req.nlh.nlmsg_len = sizeof req;
  req.nlh.nlmsg_type = type;
  req.nlh.nlmsg_flags = NLM_F_ROOT | NLM_F_MATCH | NLM_F_REQUEST;
  req.nlh.nlmsg_pid = 0;
  req.nlh.nlmsg_seq = ++netlink.seq;
  req.g.rtgen_family = family;
  
  ret = sendto (netlink.sock, (void*) &req, sizeof req, 0, 
                (struct sockaddr*) &snl, sizeof snl);
  if (ret < 0)
    {
      perror ("netlink sendto failed");
      exit (-1);
    }
  return 0;
}

/* Recieve message from netlink interface and pass those information
   to the given function. */
int
netlink_parse_info (int (*filter) (struct sockaddr_nl *, struct nlmsghdr *))
{
  int status;
  int ret;
  int seq = 0;

  while (1)
    {
      char buf[4096];
      struct iovec iov = { buf, sizeof buf };
      struct sockaddr_nl snl;
      struct msghdr msg = { (void*)&snl, sizeof snl, &iov, 1, NULL, 0, 0};
      struct nlmsghdr *h;

      status = recvmsg (netlink.sock, &msg, 0);

      if (status < 0)
	{
	  if (errno == EINTR)
	    continue;
	  if (errno == EWOULDBLOCK)
            return 0;
	  printf("netlink recvmsg overrun");
	  continue;
	}

      if (status == 0)
	{
	  printf("netlink EOF");
	  return -1;
	}

      if (msg.msg_namelen != sizeof snl)
	{
	  perror( "netlink sender address length error: length %d");
	  exit (-1);
	}

      for (h = (struct nlmsghdr *) buf; NLMSG_OK (h, status); 
	   h = NLMSG_NEXT (h, status))
	{
	  /* Message sequence. */
	  seq = h->nlmsg_seq;
	  
#if 0
	  /* pid and seq check. */
	  if (seq && seq != netlink.seq)
	    continue;
#endif

	  /* Finish of reading. */
	  if (h->nlmsg_type == NLMSG_DONE)
	    return 0;

	  /* Error handling. */
	  if (h->nlmsg_type == NLMSG_ERROR)
	    {
	      struct nlmsgerr *err = (struct nlmsgerr *) NLMSG_DATA (h);
	      if (h->nlmsg_len < NLMSG_LENGTH (sizeof (struct nlmsgerr)))
		printf( "netlink error: message truncated");
	      else
		printf( "netlink error: %s", 
		      strerror (-err->error));
	      return -1;
	    }

	  /* OK we got netlink message. */
	  ret = (*filter) (&snl, h);
	  if (ret < 0)
	    {
	      printf( "netlink filter function error");
	      return ret;
	    }
	}

      /* After error care. */
      if (msg.msg_flags & MSG_TRUNC)
	{
	  printf("netlink error: message truncated");
	  continue;
	}
      if (status)
	{
	  printf("netlink error: data remnant size %d", status);
	  return -1;
	}
      /* This message will be from kernel but reply to user request. */
      if (seq == 0)
	return 0;
    }
  return 0;
}

int if_idx_nexthop(struct in_addr *d)
{
struct sockaddr_nl snl;
int ret;
   
  netlink.sock = socket (AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
  if (netlink.sock < 0) {
    perror("Cannot open netlink socket");
    return -1;
  }
  
  bzero (&snl, sizeof snl);
  snl.nl_family = AF_NETLINK;
  snl.nl_groups = 0;

  /* snl.nl_groups = RTMGRP_LINK|RTMGRP_IPV4_ROUTE|RTMGRP_IPV4_IFADDR; */

 /* Bind the socket to the netlink structure for anything. */
  ret = bind (netlink.sock, (struct sockaddr *) &snl, sizeof snl);
  if (ret < 0)
    {
      perror("bind");
      close (netlink.sock);
      netlink.sock = -1;
      exit (-1);
    }
    
  /* Assign our global vars for in our netlink filters */
  match.s_addr = d->s_addr;
  best_route.idx = if_index = -1;
  ret = netlink_request (AF_INET, RTM_GETROUTE);
  ret = netlink_parse_info (netlink_routing_table);
  close (netlink.sock);
  return if_index;
}

#if 0
void
main()
{
  struct in_addr d;
  int idx;
  bzero (&match, sizeof match);
  d.s_addr = inet_addr("130.238.127.1");
  idx = if_idx_nexthop(&d);

  if(best_route.idx < 0) 
    printf("No route found\n");
  else {
    if(best_route.mlen == 0) printf(" Default ");
    else printf(" %x/%d ", best_route.dst, best_route.mlen);
    if(best_route.gw > 0) printf("Gateway = %x", best_route.gw);
    else printf(" Local route");
    printf(" Index=%d\n", best_route.idx);
  }
  printf("\n");
}
#endif 


