/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include "pim_common.h"
ZBUG_EXTERN


char b1[16], b2[16], b3[16], b4[16];  /* For inet_2a */

extern int pim_sock;
extern int in_cksum (void *ptr, int nbytes);
extern list pim_neighbor_list;
extern list mrt_hash[];
extern list mgrp_hash[];
extern list msrc_hash[MSRC_HASH_SIZE];
extern struct best_route best_route;
extern int pim_sock, igmp_sock;

extern char pim_sendbuf[];


void parse_prune_rt_1_wc_1(u_int16_t holdtime, struct encoded_group *eg, 
		     struct encoded_source *es, struct interface *ifp)
{
  struct mgrp *mg;
  struct oif *mi;
  listnode node;
  time_t timer;
  
  if((mg = mg_lookup( eg->addr )) == NULL ) {
    printf("Warning prune message contained non-existent group %s\n",
	   inet_2a(eg->addr, b1));
    return ;
  }

  for (node = listhead (mg->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp ) {
      if( mi->pim.t_oif_timeout) thread_cancel(mi->pim.t_oif_timeout);
      if( ifp->flags & IFF_POINTOPOINT)
	timer = 0;
      else
	timer = mi->pim.oif_deletion_delay;
      mi->pim.exptime = time(NULL)+ timer;
      mi->pim.t_oif_timeout = thread_add_timer(master, mg_oif_timeout, mi, 
					       timer);
      return;
    }
  }
  printf("Warning prune message for non-existent group %s from %s\n",
	 inet_2a(eg->addr, b1), ifp->name);
  return;
}


struct oif *update_mg_oif_list(struct mgrp *mg,
			       struct interface *ifp,
			       u_int16_t newtime,
			       u_int32_t *create)
{
  struct oif *mi;
  listnode node;
  
  *create = 1;
  for (node = listhead (mg->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp ) {
	*create = 0;
	break;
    }
  }
  if( *create) {
    mi = oif_alloc ( MTYPE_PIM_OIF );
    zassert(mi, -1);
    mi->ifp = ifp;
    mi->uptime = time(NULL);
    list_add_node(mg->oif_list, mi);
    mi->mg = mg;
    mi->uptime = time(NULL);
    mi->pim.exptime = 0; /* Set correct holdtime below */
  }

  mi->pim.oif_deletion_delay = newtime/3;
  mi->flags |= MRT_PIM_IN_USE;  
  mg->flags &= ~MRT_PRUNED;
  
  if(mi->pim.exptime < time(NULL) + newtime) {
    if( mi->pim.t_oif_timeout) thread_cancel(mi->pim.t_oif_timeout);
    mi->pim.exptime = time(NULL)+newtime;
    mi->pim.oif_deletion_delay = newtime/3;
    mi->pim.t_oif_timeout = thread_add_timer(master, mg_oif_timeout, mi, 
					 newtime);
  }
  else   printf("Ignoring oif timer setting\n");
  mg_pim_timer_update(mg, newtime);
  return mi;
}




void update_join_rt_0_wc_0(struct mrt *mr, struct interface *ifp,
			   u_int32_t we_are_lasthop)
{
  vifbitmap_t oifs;
  u_int32_t iif;


  /* We are in SPT state, lets check if we need to add a lost
     interface */

  if(mr->flags & MRT_SPT)
    {
      if(!mr_oif_lookup(mr, ifp))
	{
	  mr_add_oif(mr, ifp);
	  iif = mr_iif( mr );
	  oifs = mr_oifs(mr);
    
	  mfc_add(igmp_sock,
		  (u_int32_t) mr->ms->src,
		  (u_int32_t) mr->mg->group,
		  iif, 
		  oifs);
	  ZBUG_G(mr->mg->group)
	    {
	      fprintf(ZBUG_LOG, "Reactivated SPT oif (%s,%s)\n",
		      inet_2a(mr->ms->src, b1),
		      inet_2a(mr->mg->group, b2));
	      fflush(ZBUG_LOG);
	    }
	}
      return;
    }

  /* We already got (S,G) and if we are last hop. 
     We start transition to SPT-tree here */
    
  if( (mr->flags & MRT_RPT) && we_are_lasthop ) {
    printf("LASTHOP %s\n", inet_2a(mr->ms->src, b1));
    
    ZBUG_G(mr->mg->group)
      {
	fprintf(ZBUG_LOG, "Switch to SPT (%s,%s)\n",
	      inet_2a(mr->ms->src, b1),
		inet_2a(mr->mg->group, b2));
	fflush(ZBUG_LOG);
      }

    mr->flags |= MRT_SPT;
    
    if(!mr_oif_lookup(mr, ifp))
      mr_add_oif(mr, ifp);
    
    iif = mr_iif( mr );
    mr_oif_delete(mr, mr->ms->ifp);
    oifs = mr_oifs(mr);
    
    mfc_add(igmp_sock,
	    (u_int32_t) mr->ms->src,
	    (u_int32_t) mr->mg->group,
	    iif, 
	    oifs);
  }
}


void parse_join_rt_0_wc_0(u_int16_t holdtime, struct encoded_group *eg, 
			  struct encoded_source *es, struct interface *ifp)
{
  struct mgrp *mg;
  struct mrt *mr;
  struct msrc *ms;
  u_int32_t we_are_lasthop = 0;
 
  printf("RX join for (%s, %s) holdtime=%d\n",
	   inet_2a(es->addr, b1), inet_2a(eg->addr, b2), 
	   holdtime );


  /* Seems like the at least our Ciscos don't time-out old SPT-trees
     for a very long time. It seems like a bug :-) This causes a lot
     of unnecessary traffic load and plus routes/states. */

  /* Special case we got (S,G) join and have no (*,G) to avoid sending
     (*,G) joins we blackhole until state changes. */

  if((mg = mg_lookup(eg->addr)) == NULL ) {
    return ;
  }

  if(mg->flags & MRT_BLACKHOLE) return;

  if(!(mg->flags & MRT_PIM_IN_USE))
    {
      /* Activate *,G for PIM usage */
      
      mg->flags &= ~MRT_BLACKHOLE; 
      mg->flags |= (MRT_SPARSE|MRT_WC|MRT_RPT|MRT_PIM_IN_USE);
  
      mg->pim.exptime = time(NULL)+holdtime;
      if( mg->pim.t_entry_timeout )
	thread_cancel(mg->pim.t_entry_timeout);
      mg->pim.t_entry_timeout = thread_add_timer(master, rt_1_wc_1_timeout, 
						 mg, holdtime);
      sync_mg_family(mg);
    }

  /* A known source? */

  if((ms = ms_lookup(es->addr )) == NULL ) {
    ms = ms_add( es->addr);
    if(!ms) return;
  }

  if(ms->rpf_neighbor == 0 ) we_are_lasthop = 1;

  mr = mr_lookup(es->addr, eg->addr);
  if(! mr)
    {
      /* Create and prepare a fresh S,G entry */
      mr = mr_add(ms, mg);
      copy_mg_oifs(mr, mg);
      if(we_are_lasthop)
	printf("LASTHOP NEW %s\n", inet_2a(es->addr, b1));
      mr_mfc_add( mr ); /* add this new route */
    }
  
  update_join_rt_0_wc_0(mr, ifp, we_are_lasthop);
  mr->flags &= ~MRT_RPT; /* = Will be sent as (S,G) in JP */

  /* (S,G) join for source that lies on route towards on RP
     In this case we go to SPT directly */
  
  if(mr->ms->ifp == mg->pim.rp->ifp)
    mr->flags |= MRT_SPT;

  if(!we_are_lasthop)
    trigger_join_prune(ms->ifp); /* trigger towards source */
  
  oif_timer_update(mr, ifp, holdtime);
  mr_entry_timer_update(mr, holdtime);
  
  mr->flags &= ~MRT_RPT;
  mr->flags |= (MRT_PIM_IN_USE|MRT_SPARSE);
  mr->flags &= ~MRT_BLACKHOLE; 

  mg->flags |= (MRT_PIM_IN_USE|MRT_SPARSE);
  mg->flags &= ~MRT_BLACKHOLE; 

    /* 
       if(S,G) copy oifs from (S,G)SPT; break
	else if(*,G) copy oifs from (S,G)SPT; break
    */
 
}


void parse_prune_rt_0_wc_0(u_int16_t holdtime, struct encoded_group *eg, 
			   struct encoded_source *es, struct interface *ifp)
{
  struct mgrp *mg;
  struct mrt *mr;
  struct oif *mi;
  listnode node;
  time_t timer;
  
  if((mg = mg_lookup( eg->addr )) == NULL ) {
    printf("Warning prune message contained non-existent group %s\n",
	   inet_2a(eg->addr, b1));
    return ;
  }

  mr = mr_lookup(es->addr, eg->addr);
  if(! mr) {
    printf("Warning prune message for non-existent entry %s\n",
	   inet_2a(eg->addr, b1));
    return ;
  }

    
  if(mr->flags & MRT_RPT) return;

  for (node = listhead (mr->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp ) {
      if( mi->pim.t_oif_timeout) thread_cancel(mi->pim.t_oif_timeout);
      if( ifp->flags & IFF_POINTOPOINT)
	timer = 0;
      else
	timer = mi->pim.oif_deletion_delay;
      mi->pim.exptime = time(NULL)+ timer;
      mi->pim.t_oif_timeout = thread_add_timer(master, mr_oif_timeout, mi, 
					       timer);
      return;
    }
  }
  printf("Warning prune message for non-existent group %s from %s\n",
	 inet_2a(eg->addr, b1), ifp->name);
  return;
}



void parse_join_rt_1_wc_1(u_int16_t holdtime, struct encoded_group *eg, 
			  struct encoded_source *es, struct interface *ifp)
{
  struct mgrp *mg;
  struct mrt *mr;
  struct oif *mi;
  listnode node;
  int created;

  u_int32_t rp = select_rp(eg->addr);

  if( rp != es->addr ) {
    printf("Illegal rp for join. (*, %s) RP=%s our RP=%s\n",
	   inet_2a(eg->addr, b2), inet_2a(es->addr, b3), inet_2a(rp, b1));

    /* We are allowed to set our RP here but for now we just return */
    return;
  } 

  printf("RX join for (*, %s) SWR RP=%s holdtime=%d\n",
	   inet_2a(eg->addr, b1), inet_2a(es->addr, b2), 
	   holdtime );

  if((mg=mg_lookup(eg->addr)) == NULL ) {
    mg = pim_mg_add(eg->addr, holdtime);
    trigger_join_prune(mg->pim.rp->ifp); /* Is delayed. oif added later */
  }
  if(!(mg->flags & MRT_PIM_IN_USE))
    {
      /* Activate *,G for PIM usage */
      
      mg->flags |= MRT_PIM_IN_USE;
      mg->flags &= ~MRT_BLACKHOLE; 
      mg->flags |= (MRT_SPARSE|MRT_WC|MRT_RPT);
  
      mg->pim.exptime = time(NULL)+holdtime;
      if( mg->pim.t_entry_timeout )
	thread_cancel(mg->pim.t_entry_timeout);
      mg->pim.t_entry_timeout = thread_add_timer(master, rt_1_wc_1_timeout, mg, 
					     holdtime);
      
      sync_mg_family(mg);
    }

  /* add interface to (*,G) if a new interface. Update timers */
  mi = update_mg_oif_list(mg, ifp, holdtime, &created);

  /* If this was a new interface add it to each (S,G) */
  if(created)
    add_oif_to_mr(mg, ifp);

  /* Update all timers on each (S,G) for this ifp */
  for (node = listhead (mg->mrt_list); node; nextnode (node)) {
      mr = getdata (node);
#if 0
      /* (c) i . According to spec we should do this but we dont..*/
      if( in_prune_list( mr->src) ) continue;
#endif

      update_mr_oif_timers(mr, ifp, holdtime);
      if(created)
	/* Set entry timer to included holdtime when created off join/prune  */
	mr_entry_timer_update(mr, holdtime);
      else mr_entry_timer_update(mr, holdtime/3);
      /* TODO: Case d. (*,G) with SPT-bit set cleared pp 16. */
  }
}


void parse_prune_rt_1_wc_0(u_int16_t holdtime, struct encoded_group *eg, 
			   struct encoded_source *es, struct interface *ifp)
{
  struct mgrp *mg;
  struct mrt *mr;
  struct oif *mi;
  struct msrc *ms;
  vifbitmap_t oifs;
  listnode node;
  time_t timer;
  u_int32_t iif;


  /* Case 2 pp 17 */
  
  if((mg = mg_lookup( eg->addr )) == NULL ) {
    printf("Warning prune message contained non-existent group %s\n",
	   inet_2a(eg->addr, b1));
    return ;
  }


  if((ms = ms_lookup(es->addr )) == NULL ) {
    ms = ms_add( es->addr);
    if(!ms) return;
  }

  /* Case 2.c first. We have to create a (S,G) RPT-entry with I pruned */
  
  mr = mr_lookup(es->addr, eg->addr);
  if(! mr) {
    
    mr = mr_add(ms, mg);
    mr->flags |= 
     (MRT_SPARSE|MRT_REGISTER|MRT_RPT|MRT_PIM_IN_USE); 

    copy_mg_oifs(mr, mg);    
    
    iif = mg_iif( mg );
//mg->pim.rp->ifp->ifindex;
    mr_oif_delete(mr, ifp);

    oifs = mr_oifs(mr);
    mfc_add(igmp_sock,
            (u_int32_t) mr->ms->src,
            (u_int32_t) mr->mg->group,
            iif, 
            oifs);

    oif_timer_update(mr, ifp, holdtime); /* Guess ee have to set a timer */
    mr_entry_timer_update(mr, holdtime);
    
    mg->flags |= (MRT_PIM_IN_USE|MRT_SPARSE);
    mg->flags &= ~MRT_BLACKHOLE; 

    trigger_join_prune(ms->ifp); /* trigger towards source */
    return ;
  }

  /* A very messy part of the specs is case a/b/d */
  /* As a aproximation we do the following */
  

  for (node = listhead (mr->oif_list); node; nextnode (node)) {
    mi = getdata (node);
    zassert(mi, -1);
    if(mi->ifp == ifp ) {
      if( mi->pim.t_oif_timeout) thread_cancel(mi->pim.t_oif_timeout);
      if( ifp->flags & IFF_POINTOPOINT) 
	timer = 0;
      else 
	timer = mi->pim.oif_deletion_delay;
      mi->pim.exptime = time(NULL)+ timer;
      mi->pim.t_oif_timeout = thread_add_timer(master, mr_oif_timeout, mi, 
					       timer);
      mr_entry_timer_update(mr, holdtime);
      return;
    }
  }
}

#define JP_GROUP_MAX_SIZE 4096
  unsigned char jp_group_block_buf[JP_GROUP_MAX_SIZE];

void *encode_jp_group_block(struct mgrp *mg, struct pim_neighbor *pn, 
			    u_int32_t *len) 
{

  unsigned char *p = jp_group_block_buf;
  struct pim_interface *pi = pn->ifp->info;
  void *num_join_p, *num_prune_p;
  u_int16_t num_join, num_prune;
  u_int8_t bits;
  listnode node;
  struct mrt *mr;
  u_int32_t group_added;
  u_int32_t rpf_neighbor;

  *len = 0;
  num_join = num_prune = 0;
  group_added = 0;

  rpf_neighbor = mg_rpf_neighbor( mg );
  
  if(rpf_neighbor == pn->address.s_addr ) {

    /* Are we DR or have non-NULL oif? */

    if(( pi->flags & IF_DR ) ||
       ( mg->oif_list && !list_isempty(mg->oif_list) )) 
      {
	
	if( ! group_added ) {
	  
	  /* init_jp_block */
	  p += encode_group(p, mg->group, 32);
	  num_join_p = p;
	  p += 2;
	  num_prune_p = p;
	  p += 2;
	  group_added = 1;
	}
	bits = (S_BIT|WC_BIT|RPT_BIT);
	p += encode_source(p, mg->pim.rp->src, 32, bits);  
	num_join++;
  
	printf("Adding join for (*, %s) Flags=SWR RP=%s RPF=%s\n",
	       inet_2a(mg->group, b1), inet_2a(mg->pim.rp->src, b2), 
	       inet_2a(mg->pim.rp->rpf_neighbor, b3));
      }
  }


  /* Loop over all S for this G */
  for (node = listhead (mg->mrt_list); node; nextnode (node)) {
    mr = getdata(node);
    
    if(!(mr->flags & MRT_SPARSE))
      continue;
    if(mr->flags & MRT_REGISTER)
      continue;

    if(mr->flags & MRT_RPT)
      continue;

    /*
      if(mr->flags & MRT_PRUNED)
      continue;
      */

    rpf_neighbor = mr_rpf_neighbor( mr );
    
    if(rpf_neighbor == pn->address.s_addr) {

      if( ! group_added ) {
	
	/* init_jp_block */
	p += encode_group(p, mg->group, 32);
	num_join_p = p;
	p += 2;
	num_prune_p = p;
	p += 2;
	group_added = 1;
      }
      bits = (S_BIT);
      p += encode_source(p, mr->ms->src, 32, bits);  
      num_join++;

      printf("Adding join for (%s, %s) Flags=S RPF=%s\n",
	     inet_2a(mr->ms->src, b1), inet_2a(mg->group, b2), 
	     inet_2a(mr->ms->rpf_neighbor, b3));
    }
  }

  /* PRUNES ------------ */

  rpf_neighbor = mg_rpf_neighbor( mg );
    
  if(rpf_neighbor == pn->address.s_addr ) {


    /* Add prunes RPT=1, WC=1 */
    
    if( (pi->flags & IF_DR) && (mg->flags & MRT_PRUNED))
      {
	if( ! group_added ) {
	  
	  /* init_jp_block */
	  p += encode_group(p, mg->group, 32);
	  num_join_p = p;
	  p += 2;
	  num_prune_p = p;
	  p += 2;
	  group_added = 1;
	}
	bits = (S_BIT|WC_BIT|RPT_BIT);
	p += encode_source(p, mg->pim.rp->src, 32, bits);  
	num_prune++;
	
	printf("Adding prune for (*, %s) SWR RP=%s RPF=%s\n",
	       inet_2a(mg->group, b1), inet_2a(mg->pim.rp->src, b2), 
	       inet_2a(mg->pim.rp->rpf_neighbor, b3));
      }
  }
  
  /* Loop over all S for this G */

  for (node = listhead (mg->mrt_list); node; nextnode (node)) {
    int prune_4A;
    mr = getdata(node);
    
    /* Add prunes RPT=0, WC=0 Case. 3.A pp 13 */
    
    /* 3.a */

    rpf_neighbor = mr_rpf_neighbor( mr );
    
    if(rpf_neighbor == pn->address.s_addr) {    
      if( (!(mr->flags & MRT_RPT)) && (mr->flags & MRT_PRUNED))
	{
	  /* This code never runs because we never set the PRUNED
	     on any mr */
	  if( ! group_added ) {
	    
	    /* init_jp_block */
	    p += encode_group(p, mg->group, 32);
	    num_join_p = p;
	    p += 2;
	    num_prune_p = p;
	    p += 2;
	    group_added = 1;
	  }
	  bits = (S_BIT);
	  p += encode_source(p, mr->ms->src, 32, bits);  
	  num_prune++;
	  
	  printf("Adding prune for (%s, %s) FLAGS=S RPF=%s\n",
		 inet_2a(mg->group, b1), inet_2a(mr->ms->src, b2), 
		 inet_2a(mr->ms->rpf_neighbor, b3));
	}
    }


    rpf_neighbor = mg_rpf_neighbor( mg );

    if(rpf_neighbor == pn->address.s_addr ) {

      /* Add prunes RPT=1, WC=0 Case. 4.A pp 13 */
  
      /* 4.a */
      prune_4A = 0;
  
      if( (mr->flags & MRT_RPT) && (mr->flags & MRT_PRUNED)) 
	prune_4A = 1;
  
      /* 4.b */
      if(!(mr->flags & MRT_RPT) && (mr->flags & MRT_SPT) &&
	 (mr->ms->ifp != mg->pim.rp->ifp)) 
	prune_4A = 1;
  
      /* 4.c We are lasthop */
      if(mr->ms->rpf_neighbor == 0 ) prune_4A = 1;
  
      if( prune_4A ) {
	
	if( ! group_added ) {
	  
	  /* init_jp_block */
	  p += encode_group(p, mg->group, 32);
	  num_join_p = p;
	  p += 2;
	  num_prune_p = p;
	  p += 2;
	  group_added = 1;
	}
	bits = (S_BIT|RPT_BIT);
	p += encode_source(p, mr->ms->src, 32, bits);  
	num_prune++;
    
	printf("Adding prune for (%s, %s) FLAGS=SR RP=%s RPF=%s\n",
	       inet_2a(mg->group, b1), inet_2a(mr->ms->src, b2), 
	       inet_2a(mg->pim.rp->src, b3),
	       inet_2a(mg->pim.rp->rpf_neighbor, b4));
      }
    }
  }

  if( num_join + num_prune == 0) {
    *len = 0;
    return (void *) NULL;
  }

  num_join = htons(num_join);
  memcpy(num_join_p, &num_join, 2);
  num_prune = htons(num_prune);
  memcpy(num_prune_p, &num_prune, 2);

  *len = p - jp_group_block_buf;
  return (void *) jp_group_block_buf;
}


void send_join_prune(struct interface *ifp)
{
  struct pim_neighbor *pn;
  struct pim_interface *pi;
  struct mgrp *mg;
  int num_groups;
  listnode node, node1;
  void *p, *buf_p;
  unsigned char *numgroups_p;
  u_int32_t len, idx;
  u_int16_t holdtime;
  u_int32_t jplen = 0;
  
  /* For all pim neigbors */
  for (node = listhead (pim_neighbor_list); node; nextnode (node)) {
    pn = getdata(node);
    if(ifp != pn->ifp ) continue; /* Not this interface */
    pi=pn->ifp->info;

    /* init_jp_message */
    
    num_groups = 0; /* Fill in later */
    buf_p = (char *) (pim_sendbuf + sizeof(struct ip)  + 
		      sizeof(struct pimheader));
    
    buf_p += encode_unicast(buf_p, pn->address.s_addr);
    *(u_int8_t *)buf_p = 0; /* Reserved */
    buf_p += 1;
    
    numgroups_p = buf_p++; /* Fill in later */
    holdtime = pi->join_prune_holdtime;
    holdtime = htons(holdtime);
    memcpy(buf_p, &holdtime, 2);
    buf_p += 2;
    
    /* Scan select G from mgrp */
    for(idx=0; idx < MGRP_HASH_SIZE; idx++) {
      if( mgrp_hash[idx] == 0)
	continue;
      
      for(node1=listhead (mgrp_hash[idx]); node1; nextnode (node1)) {
	mg = getdata( node1 );

	p =  encode_jp_group_block(mg, pn, &len) ;
	if( len ) {
	  memcpy( buf_p, p, len);
	  buf_p += len;
	  jplen += len;
	  num_groups++;
	}
      }
    }
    if(num_groups == 0) return;
    jplen +=  sizeof(struct encoded_unicast) + 4; /* The JP header size */
    *numgroups_p = (unsigned char )num_groups; 
    if(pi->version == 2)
      send_pim_v2_packet(pi->ifp, pim_sendbuf, jplen, PIM_JOIN_PRUNE,
			 htonl(INADDR_PIM_ALLRTRS_GROUP), 1);  
  }
}


int send_join_prune_thread(struct thread *t_send_join_prune)
{

  struct pim_interface *pi = THREAD_ARG (t_send_join_prune);
  struct interface *ifp;
  ifp = pi->ifp;

  send_join_prune(ifp);
  pi->t_send_join_prune = thread_add_timer(master, send_join_prune_thread, 
				 pi, pi->join_prune_query_interval);
  return 0;
}


/* Trigger join/prunes introduces a short delay before sending the
   the triggered join/prune message. With this we hope not to send
   a join prune for every triggered event. */

int trigger_join_prune_thread(struct thread *t_trigger_join_prune)
{
  struct pim_interface *pi = THREAD_ARG (t_trigger_join_prune);
  struct interface *ifp;
  ifp = pi->ifp;
  pi->trigger_join_prune_active=0;

  printf("*********** SENDING TRIGGER JOIN PRUNE\n");
  send_join_prune(ifp);
  return 0;
}

void trigger_join_prune(struct interface *ifp)
{
  struct pim_interface *pi;
  pi = ifp->info;
  
  printf("*********** TRIGGER JOIN PRUNE\n");
  if(pi->trigger_join_prune_active) return; /* We already have a timer */
  pi->trigger_join_prune_active=1;

  thread_add_timer(master, trigger_join_prune_thread, pi, 
					   pi->trigger_join_prune_delay);
  return;
}






